#include "StdAfx.h"
#include "JSPluginManager.h"
#include "../js/JSValueWrapper.h"
#include "../window/IWindow.h"
#include "../js/JSDispatch.h"
#include "../js/IJSContext.h"
#include "../ar/ARWindow.h"
#include "../common/URLPath.h"

#define FUNC_LOAD "load"
#define PROP_PLUGIN_COUNT	"pluginCount"

#define DLL_FUNCTION_CREATE_PLUGIN "_CreatePlugin@8"

// \z܂B
JSPluginManager::JSPluginManager(const char* basePath, IWindow* pWindow, ARWindow* pARWindow)
: JSMemberCollection("JSPluginManager")
{
	m_basePath = basePath;
	m_pWindow = pWindow;
	m_pARWindow = pARWindow;
	m_pJSEngineContext = NULL;
	m_pEnv = new ARGPluginEnvironmentImpl(this);
}

// j܂B
JSPluginManager::~JSPluginManager()
{
	for(vector<IARGPlugin*>::size_type i = 0; i < m_plugins.size(); i ++) {
		IARGPlugin* pPlugin = m_plugins[i].second;
		if(pPlugin != NULL) {
			pPlugin->Detach();
			pPlugin->Release();
			pPlugin = NULL;
		}

		HMODULE hModule = m_plugins[i].first;
		if(hModule != NULL) {
			FreeLibrary(hModule);
			hModule = NULL;
		}

		m_plugins[i].first = NULL;
		m_plugins[i].second = NULL;
	}
	m_plugins.clear();
	if(m_pEnv != NULL) {
		m_pEnv->Release();
		m_pEnv = NULL;
	}
}

// j܂B
void JSPluginManager::Destroy()
{
	delete this;
}

// JavaScriptGWݒ肵܂B
HRESULT JSPluginManager::SetJSEngineContext(JSEngineContext* pContext)
{
	if(pContext == NULL) {
		return E_POINTER;
	}
	m_pJSEngineContext = pContext;
	return S_OK;
}

// o܂B
HRESULT JSPluginManager::InitMembers()
{
	HRESULT hr = S_OK;

	hr = DefineFunction(FUNC_LOAD, 2);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty(PROP_PLUGIN_COUNT, true);
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

typedef HRESULT (__stdcall *CreatePluginFunc) (const char*, IUnknown** ppUnk);

// ֐s܂B
HRESULT JSPluginManager::InvokeFunction(JSContext *cx, JSObject *obj, const char* name, uintN argc, jsval *argv, jsval *rval)
{
	if(cx == NULL || obj == NULL || name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == FUNC_LOAD) {
		if(argc < 2) {
			return E_INVALIDARG;
		}
		
		string path;
		JSValueWrapper value1(cx, argv[0]);
		HRESULT hr = value1.GetString(path);
		if(FAILED(hr)) {
			return hr;
		}

		string name;
		JSValueWrapper value2(cx, argv[1]);
		hr = value2.GetString(name);
		if(FAILED(hr)) {
			return hr;
		}

		string fullpath = m_basePath + path;
		HMODULE hModule = LoadLibraryA(fullpath.c_str());
		if(hModule == NULL) {
			return E_UNEXPECTED;
		}

		CreatePluginFunc pFunc = NULL;
		pFunc = (CreatePluginFunc) GetProcAddress(hModule, DLL_FUNCTION_CREATE_PLUGIN);
		if(pFunc == NULL) {
			return E_UNEXPECTED;
		}

		IUnknown* pUnk = NULL;
		hr = pFunc(name.c_str(), & pUnk);
		if(FAILED(hr)) {
			FreeLibrary(hModule);
			hModule = NULL;
			return hr;
		}
		if(pUnk == NULL) {
			FreeLibrary(hModule);
			hModule = NULL;
			return E_UNEXPECTED;
		}

		IARGPlugin* pPlugin = NULL;
		hr = pUnk->QueryInterface(IID_IARGPlugin, (void**) & pPlugin);
		if(FAILED(hr)) {
			pUnk->Release();
			pUnk = NULL;
			FreeLibrary(hModule);
			hModule = NULL;
			return hr;
		}
		pUnk->Release();
		pUnk = NULL;

		hr = pPlugin->Attach(m_pEnv);
		if(FAILED(hr)) {
			pPlugin->Release();
			pPlugin = NULL;
			FreeLibrary(hModule);
			hModule = NULL;
			return hr;
		}

		pair<HMODULE,IARGPlugin*> loaded(hModule, pPlugin);
		m_plugins.push_back(loaded);

		return S_OK;
	}
	return JSMemberCollection::InvokeFunction(cx, obj, name, argc, argv, rval);
}

// vpeB擾܂B
HRESULT JSPluginManager::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	if(cx == NULL || obj == NULL || name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == PROP_PLUGIN_COUNT) {
		JSValueWrapper value(cx, JSVAL_NULL);
		HRESULT hr = value.SetInt(m_plugins.size());
		if(FAILED(hr)) {
			return hr;
		}

		return value.Get(vp);
	}
	return JSMemberCollection::GetProperty(cx, obj, name, vp);
}

// \z܂B
JSPluginManager::ARGPluginEnvironmentImpl::ARGPluginEnvironmentImpl(JSPluginManager* pOwner)
{
	m_pOwner = pOwner;
}

// j܂B
JSPluginManager::ARGPluginEnvironmentImpl::~ARGPluginEnvironmentImpl()
{
	m_pOwner = NULL;
}

// pX擾܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::GetRealPath(const char* relativePath, char* absolutePath, int maxLen, int* pLen)
{
	if(relativePath == NULL || pLen == NULL) {
		return E_POINTER;
	}

	string url;
	HRESULT hr = m_pOwner->m_pContext->GetAbsoluteURL(relativePath, url);
	if(FAILED(hr)) {
		return hr;
	}

	URLPath urlPath(url.c_str(), & hr);
	if(FAILED(hr)) {
		return hr;
	}
	string realFile;
	hr = urlPath.GetRealPath(realFile);
	if(FAILED(hr)) {
		return hr;
	}

	if(absolutePath != NULL) {
		strcpy_s(absolutePath, maxLen, realFile.c_str());
		if(realFile.length() < maxLen) {
			absolutePath[realFile.length()] = '\0';
		}
	}

	*pLen = realFile.length();

	return S_OK;
}

// vpeBǉ܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::AddProperty(const char* name, IDispatch* pDisp)
{
	if(name == NULL || pDisp == NULL) {
		return E_POINTER;
	}
	if(m_pOwner->m_pJSEngineContext == NULL) {
		return E_UNEXPECTED;
	}

	HRESULT hr = S_OK;
	JSDispatch* pJSDisp = new JSDispatch(pDisp, & hr);
	if(FAILED(hr)) {
		delete pJSDisp;
		return hr;
	}

	hr = pJSDisp->Attach(m_pOwner->m_pJSEngineContext, NULL);
	if(FAILED(hr)) {
		return hr;
	}
	return m_pOwner->m_pJSEngineContext->DefineGlobalProperty(name, pJSDisp);
}

// IuWFNgt@Ngǉ܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::AddObjectFactory(const char* name, IARGObjectFactory* pObjectFactory)
{
	if(name == NULL || pObjectFactory == NULL) {
		return E_POINTER;
	}
	return m_pOwner->m_pARWindow->AddObjectFactory(name, pObjectFactory);
}

// R\[ɕo͂܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::PrintConsole(const char* text)
{
	if(text == NULL) {
		return E_POINTER;
	}
	return m_pOwner->m_pWindow->PrintLine(text);
}
