// -*- Mode: C++; indent-tabs-mode: nil; c-basic-offset: 4; -*-
//
// OpenVRML
//
// Copyright (C) 1998  Chris Morley
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

# ifndef OPENVRML_BROWSER_H
#   define OPENVRML_BROWSER_H

#   include <cstddef>
#   include <list>
#   include <map>
#   include <openvrml/common.h>
#   include <openvrml/node_class_ptr.h>
#   include <openvrml/script.h>
# include "scope.h"

namespace openvrml {

    class invalid_vrml : public std::runtime_error {
    public:
        invalid_vrml();
        virtual ~invalid_vrml() throw ();
    };

    class viewer;
    class ProtoNode;
    class scene;
    class Vrml97RootScope;
    class null_node_class;
    class null_node_type;

    class browser {
        friend class Vrml97Parser;
        friend class ProtoNodeClass;
        friend class Vrml97RootScope;

    public:
        enum cb_reason {
            destroy_world_id,
            replace_world_id
        };

        typedef void (*scene_cb)(cb_reason reason);

    private:
        std::auto_ptr<null_node_class> null_node_class_;
        std::auto_ptr<null_node_type> null_node_type_;
        typedef std::map<std::string, node_class_ptr> node_class_map_t;
        node_class_map_t node_class_map;
        script_node_class script_node_class_;
        scene * scene_;
        node_ptr default_viewpoint;
        viewpoint_node * active_viewpoint_;
        std::list<viewpoint_node *> viewpoint_list;
        typedef std::list<node_ptr> bind_stack_t;
        bind_stack_t navigation_info_stack;
        std::list<node *> navigation_infos;
        std::list<node *> scoped_lights;
        std::list<script_node *> scripts;
        std::list<node *> timers;
        std::list<node *> audio_clips;
        std::list<node *> movies;
        std::list<ProtoNode *> proto_node_list;
        bool modified_;
        bool new_view;
        double delta_time;

    protected:
        typedef std::list<scene_cb> scene_cb_list_t;

        struct event {
            double timestamp;
            field_value * value;
            node_ptr to_node;
            std::string to_eventin;
        };

        scene_cb_list_t scene_callbacks;

        double frame_rate_;

        enum { max_events = 400 };
        event event_mem[max_events];
        size_t first_event;
        size_t last_event;

    public:
        static double current_time() throw ();

        std::ostream & out;
        std::ostream & err;
        bool flags_need_updating;

        browser(std::ostream & out, std::ostream & err) throw (std::bad_alloc);
        virtual ~browser() throw ();

        const std::vector<node_ptr> & root_nodes() const throw ();
        const node_path find_node(const node & n) const throw (std::bad_alloc);
        viewpoint_node & active_viewpoint() const throw ();
        void active_viewpoint(viewpoint_node & viewpoint) throw ();
        void reset_default_viewpoint() throw ();
        void add_viewpoint(viewpoint_node & viewpoint) throw (std::bad_alloc);
        void remove_viewpoint(viewpoint_node & viewpoint) throw ();
        const std::list<viewpoint_node *> & viewpoints() const throw ();

        virtual const char * name() const throw ();
        virtual const char * version() const throw ();
        float current_speed();
        const std::string world_url() const throw (std::bad_alloc);
        void replace_world(const std::vector<node_ptr> & nodes);
        virtual void load_url(const std::vector<std::string> & url,
                              const std::vector<std::string> & parameter)
            throw (std::bad_alloc);
        virtual void description(const std::string & description);
        const std::vector<node_ptr> create_vrml_from_stream(std::istream & in);
        void create_vrml_from_url(const std::vector<std::string> & url,
                                  const node_ptr & node,
                                  const std::string & event);

        void add_world_changed_callback(scene_cb);

        void sensitive_event(node * object, double timestamp,
                             bool is_over, bool is_active, double * point);

        void queue_event(double timestamp, field_value * value,
                         const node_ptr & toNode,
                         const std::string & to_eventin);

        bool events_pending();

        void flush_events();

        double frame_rate() const;

        bool update(double current_time = -1.0);

        void render(openvrml::viewer & viewer);

        void modified(bool value);
        bool modified() const;

        void delta(double d);
        double delta() const;

        void add_navigation_info(vrml97_node::navigation_info_node &);
        void remove_navigation_info(vrml97_node::navigation_info_node &);
        vrml97_node::navigation_info_node *bindable_navigation_info_top();
        void bindable_push(vrml97_node::navigation_info_node *);
        void bindable_remove(vrml97_node::navigation_info_node *);

        void add_scoped_light(vrml97_node::abstract_light_node &);
        void remove_scoped_light(vrml97_node::abstract_light_node &);

        void add_time_sensor(vrml97_node::time_sensor_node &);
        void remove_time_sensor(vrml97_node::time_sensor_node &);

        void add_audio_clip(vrml97_node::audio_clip_node &);
        void remove_audio_clip(vrml97_node::audio_clip_node &);

        void add_movie(vrml97_node::movie_texture_node &);
        void remove_movie(vrml97_node::movie_texture_node &);

        void add_proto(ProtoNode & node);
        void remove_proto(ProtoNode & node);
        void add_script(script_node &);
        void remove_script(script_node &);

        void update_flags();

    protected:
        bool headlight_on();
        void do_callbacks(cb_reason reason);

        const node_ptr bindable_top(const bind_stack_t & stack);
        void bindable_push(bind_stack_t & stack, const node_ptr & node);
        void bindable_remove(bind_stack_t & stack, const node_ptr & node);

    private:
        // Not copyable.
        browser(const browser &);
        browser & operator=(const browser &);

        void init_node_class_map();
    };


    class bad_url : public std::runtime_error {
    public:
        bad_url(const std::string & message);
        virtual ~bad_url() throw ();
    };


    class invalid_url : public bad_url {
    public:
        invalid_url();
        virtual ~invalid_url() throw ();
    };


    class unreachable_url : public bad_url {
    public:
        unreachable_url();
        virtual ~unreachable_url() throw ();
    };


    class scene {
        std::vector<node_ptr> nodes_;
        std::string url_;

    public:
        openvrml::browser & browser;
        scene * const parent;

        scene(openvrml::browser & browser,
              const std::vector<std::string> & url,
              scene * parent = 0)
            throw (invalid_vrml, std::bad_alloc);

        void initialize(double timestamp) throw (std::bad_alloc);
        const std::vector<node_ptr> & nodes() const throw ();
        const std::string url() const throw (std::bad_alloc);
        void render(openvrml::viewer & viewer, rendering_context context);
        void load_url(const std::vector<std::string> & url,
                      const std::vector<std::string> & parameter)
                throw (std::bad_alloc);
        void shutdown(double timestamp) throw ();

    private:
        // Noncopyable.
        scene(const scene &);
        scene & operator=(const scene &);
    };

    inline const std::vector<node_ptr> & scene::nodes() const throw()
    {
        return this->nodes_;
    }

    class NodeInterfaceTypeMismatch : public std::runtime_error {
    public:
        NodeInterfaceTypeMismatch(node_interface::type_id lhs,
                                  node_interface::type_id rhs);
        virtual ~NodeInterfaceTypeMismatch() throw ();
    };

    class ProtoNode : public node {
        friend class ProtoNodeClass;
        friend class Vrml97Parser;

    public:
        struct ImplNodeInterface {
            openvrml::node & node;
            std::string interfaceId;

            ImplNodeInterface(openvrml::node & node,
                              const std::string & interfaceId);
        };
        typedef std::multimap<std::string, ImplNodeInterface> ISMap;
        typedef std::map<std::string, field_value_ptr> FieldValueMap;
        typedef std::map<std::string, polled_eventout_value> EventOutValueMap;

        ISMap isMap;

    private:
        FieldValueMap unISdFieldValueMap;
        EventOutValueMap eventOutValueMap;
        std::vector<node_ptr> implNodes;

    public:
        explicit ProtoNode(const node_type & nodeType) throw (std::bad_alloc);
        ProtoNode(const node_type & nodeType,
                  const scope_ptr & scope,
                  const ProtoNode & node) throw (std::bad_alloc);
        virtual ~ProtoNode() throw ();

        const std::vector<node_ptr> & getImplNodes() const throw ();

        void addRootNode(const node_ptr & node) throw (std::bad_alloc);
        void addIS(node & implNode,
                   const std::string & implNodeInterfaceId,
                   const std::string & protoInterfaceId)
            throw (unsupported_interface, NodeInterfaceTypeMismatch,
                   field_value_type_mismatch, std::bad_alloc);

        void update(double time);

        virtual const script_node * to_script() const throw ();
        virtual script_node * to_script() throw ();
        virtual const appearance_node * to_appearance() const throw ();
        virtual appearance_node * to_appearance() throw ();
        virtual const child_node * to_child() const throw ();
        virtual child_node * to_child() throw ();
        virtual const color_node * to_color() const throw ();
        virtual color_node * to_color() throw ();
        virtual const coordinate_node * to_coordinate() const throw ();
        virtual coordinate_node * to_coordinate() throw ();
        virtual const font_style_node * to_font_style() const throw ();
        virtual font_style_node * to_font_style() throw () ;
        virtual const geometry_node * to_geometry() const throw ();
        virtual geometry_node * to_geometry() throw ();
        virtual const material_node * to_material() const throw ();
        virtual material_node * to_material() throw ();
        virtual const normal_node * to_normal() const throw ();
        virtual normal_node * to_normal() throw ();
        virtual const sound_source_node * to_sound_source() const throw ();
        virtual sound_source_node * to_sound_source() throw ();
        virtual const texture_node * to_texture() const throw ();
        virtual texture_node * to_texture() throw ();
        virtual const texture_coordinate_node * to_texture_coordinate() const
                throw ();
        virtual texture_coordinate_node * to_texture_coordinate() throw ();
        virtual const texture_transform_node * to_texture_transform() const
                throw ();
        virtual texture_transform_node * to_texture_transform() throw ();
        virtual const transform_node * to_transform() const throw ();
        virtual transform_node * to_transform() throw ();
        virtual const viewpoint_node * to_viewpoint() const throw ();
        virtual viewpoint_node * to_viewpoint() throw ();

        virtual vrml97_node::anchor_node * to_anchor() const;
        virtual vrml97_node::audio_clip_node * to_audio_clip() const;
        virtual vrml97_node::cylinder_sensor_node * to_cylinder_sensor() const;
        virtual vrml97_node::abstract_light_node * to_light() const;
        virtual vrml97_node::movie_texture_node * to_movie_texture() const;
        virtual vrml97_node::navigation_info_node * to_navigation_info() const;
        virtual vrml97_node::plane_sensor_node * to_plane_sensor() const;
        virtual vrml97_node::point_light_node * to_point_light() const;
        virtual vrml97_node::sphere_sensor_node * to_sphere_sensor() const;
        virtual vrml97_node::spot_light_node * to_spot_light() const;
        virtual vrml97_node::time_sensor_node * to_time_sensor() const;
        virtual vrml97_node::touch_sensor_node * to_touch_sensor() const;

        virtual void render(openvrml::viewer & viewer,
                            rendering_context context);

    private:
        // Not copyable.
        ProtoNode(const ProtoNode &);
        ProtoNode & operator=(const ProtoNode &);

        virtual void do_initialize(double timestamp) throw (std::bad_alloc);
        virtual void do_field(const std::string & id,
                                 const field_value & value)
                throw (unsupported_interface, std::bad_cast, std::bad_alloc);
        virtual const field_value & do_field(const std::string & id) const
                throw (unsupported_interface);
        virtual void do_process_event(const std::string & id,
                                      const field_value & value,
                                      double timestamp)
                throw (unsupported_interface, std::bad_cast, std::bad_alloc);
        virtual const field_value &
        do_eventout(const std::string & id) const
            throw (unsupported_interface);
        virtual void do_shutdown(double timestamp) throw ();
    };


    class ProtoNodeClass : public node_class {
        friend class ProtoNode;
        friend class Vrml97Parser;

    public:
        class ProtoNodeType : public node_type {
            node_interface_set nodeInterfaces;

        public:
            ProtoNodeType(ProtoNodeClass & nodeClass, const std::string & id)
                throw (unsupported_interface, std::bad_alloc);
            virtual ~ProtoNodeType() throw ();

            virtual const node_interface_set & interfaces() const throw ();
            virtual const node_ptr create_node(const scope_ptr & scope) const
                throw (std::bad_alloc);

            void addInterface(const node_interface & interface)
                throw (std::invalid_argument, std::bad_alloc);
        };

        friend class ProtoNodeType;

    private:
        typedef std::map<std::string, field_value_ptr> DefaultValueMap;

        ProtoNodeType protoNodeType;
        DefaultValueMap defaultValueMap;
        ProtoNode protoNode;

    public:
        explicit ProtoNodeClass(openvrml::browser & browser) throw ();
        virtual ~ProtoNodeClass() throw ();

        void addEventIn(field_value::type_id, const std::string & id)
            throw (std::invalid_argument, std::bad_alloc);
        void addEventOut(field_value::type_id, const std::string & id)
            throw (std::invalid_argument, std::bad_alloc);
        void addExposedField(const std::string & id,
                             const field_value_ptr & defaultValue)
            throw (std::invalid_argument, std::bad_alloc);
        void addField(const std::string & id,
                      const field_value_ptr & defaultValue)
            throw (std::invalid_argument, std::bad_alloc);
        void addRootNode(const node_ptr & node) throw (std::bad_alloc);
        void addIS(node & implNode,
                   const std::string & implNodeInterfaceId,
                   const std::string & protoInterfaceId)
            throw (unsupported_interface, NodeInterfaceTypeMismatch,
                   field_value_type_mismatch, std::bad_alloc);

        virtual const node_type_ptr create_type(const std::string & id,
                                              const node_interface_set &)
            throw (unsupported_interface, std::bad_alloc);
    };

    class Vrml97RootScope : public scope {
    public:
        Vrml97RootScope(const browser & browser,
                        const std::string & uri = std::string())
            throw (std::bad_alloc);
        virtual ~Vrml97RootScope() throw ();
    };
}

# endif // OPENVRML_BROWSER_H
